#include "light.h"
#include "material.h"
#include "scene.h"
#include "RTConfig.h"

#include <qmessagebox.h>

using namespace qglviewer;

void Light::initFromDOMElement(const QDomElement& e)
{
  QDomNode n = e.firstChild();
  while (!n.isNull())
    {
      QDomElement e = n.toElement();
      if (!e.isNull())
	{
	  if (e.tagName() == "Color")
	    color_.initFromDOMElement(e);
	}
      else
	QMessageBox::warning(NULL, "Light XML error", "Error while parsing Light XML document");
      n = n.nextSibling();
    }
}





/********** AMBIENT LIGHT **********/



Color AmbientLight::illuminatedColor(const qglviewer::Vec& eye, const Hit& hit) const {
	Color res;

	if(RTConfig::getInstance()->getTexEnable()){
		// "Partie XI: Textures"
		//With texture
		res=hit.material().diffuseColor(hit.u(), hit.v())*color();
	}else{
		// Without texture
		res=hit.material().diffuseColor()*color(); 
	}

	return res;
}


// Build this "real" light in the scene to check the rendering
void AmbientLight::draw() const{
	GLfloat ambient[] = { color().r, color().g, color().b };
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glEnable(GL_LIGHT0);
}


bool AmbientLight::visibleFrom(const qglviewer::Vec& pos, const Scene* const scene, Hit &rayhit) const {
	// Always visible
	return true;
}





/********** DIRECTIONAL LIGHT **********/



DirectionalLight::DirectionalLight()
{
  setDirection(Vec(1.0, 0.0, 0.0));
}


void DirectionalLight::initFromDOMElement(const QDomElement& e)
{
  Light::initFromDOMElement(e);
  
  QDomNode n = e.firstChild();
  while (!n.isNull())
    {
      QDomElement e = n.toElement();
      if (!e.isNull())
	{
	  if (e.tagName() == "Direction")
	    {
	      Vec dir;
	      dir.initFromDOMElement(e);
	      setDirection(dir);
	    }
	}
      else
	QMessageBox::warning(NULL, "DirectionalLight XML error", "Error while parsing DirectionalLight XML document");
      n = n.nextSibling();
    }
}


// Draw a line from the center of the scene to represent the direction of this light
void DirectionalLight::draw() const{
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  glLineWidth(3.0f);
  glColor4f(color().r, color().g, color().b, 1);
  glBegin(GL_LINES);
  glVertex3f(0,0,0);
  glVertex3f(direction_.x, direction_.y, direction_.z);
  glEnd();

  glPopAttrib();
  glPopMatrix();

  // Build this "real" light in the scene to check the rendering
  GLfloat diffuseLight[] = { color().r, color().g, color().g, 1.0f };
  GLfloat specularLight[] = { color().r, color().g, color().g, 1.0f };
  GLfloat position[] = { -direction_.x, -direction_.y, -direction_.z, 0.0f};

  glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuseLight);
  glLightfv(GL_LIGHT1, GL_SPECULAR, specularLight);
  glLightfv(GL_LIGHT1, GL_POSITION, position);

  glEnable(GL_LIGHT1);
}


Color DirectionalLight::illuminatedColor(const qglviewer::Vec& eye, const Hit& hit) const{
	Color res;
	
	qglviewer::Vec source = -direction_;
	source.normalize();

	float lidn = source*hit.normal();
	if(lidn<0)
		return Color(0,0,0);

	Color diffus;
	if(RTConfig::getInstance()->getTexEnable()){
		// "Partie XI: Textures"
		// With texture
		diffus = hit.material().diffuseColor(hit.u(), hit.v())*(lidn);
	}else{
		// Without texture
		diffus = hit.material().diffuseColor()*(lidn);
	}
	
	
	Color specular(0,0,0);

	//Phong
	if (RTConfig::getInstance()->getLightMode() == RTConfig::RTC_LM_PHONG) {
		//std::cout<<"pbm\n";
		
		qglviewer::Vec normale = hit.normal();
		qglviewer::Vec refl = -source + (normale*2.0f*(source*normale));

		float dp = eye*refl;
		if(dp<0)
			dp=0;

		specular = hit.material().specularColor() * pow( dp, hit.material().specularCoefficient());
	}

	//Blinn-Phong
	else {
		qglviewer::Vec h = eye+source;
		h.normalize();

		float dp = h*hit.normal();
		if(dp<0)
			dp=0;
		specular = hit.material().specularColor() * pow( dp, hit.material().specularCoefficient());
	}


	res = color() * (diffus+specular);

	return res;
}


bool DirectionalLight::visibleFrom(const qglviewer::Vec& pos, const Scene* const scene, Hit &rayhit) const{
	Ray ray;
//	Hit rayhit;

	ray.setStart(pos/*-direction_*0.01*/);
	ray.setDirection(-direction_);

	rayhit.setTime(1.0E5);
	rayhit.setIntersection(qglviewer::Vec(0,0,0));

	scene->intersect(ray, rayhit);

	// return true only if there is no intersection
	return (rayhit.time()==1.0E5);
}





/********** POINT LIGHT **********/



void PointLight::initFromDOMElement(const QDomElement& e)
{
  Light::initFromDOMElement(e);
  
  QDomNode n = e.firstChild();
  while (!n.isNull())
    {
      QDomElement e = n.toElement();
      if (!e.isNull())
	{
	  if (e.tagName() == "Position")
	    position_.initFromDOMElement(e);
	}
      else
	QMessageBox::warning(NULL, "PointLight XML error", "Error while parsing PointLight XML document");
      n = n.nextSibling();
    }
}


// Draw a sphere at the position of this point light to show its place 
void PointLight::draw() const{

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glColor4f(color().r, color().g, color().b, 1);
  
  glTranslatef(position_.x, position_.y, position_.z);

  GLUquadricObj *quad=gluNewQuadric();
  gluSphere(quad, 0.01 , 10, 10);

  glPopMatrix();

  // Build this "real" light in the scene to chech the rendering
  GLfloat diffuseLight[] = { color().r, color().g, color().g, 1.0f };
  GLfloat specularLight[] = { color().r, color().g, color().g, 1.0f };
  GLfloat position[] = { position_.x, position_.y, position_.z, 1.0f};

  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
  glLightfv(GL_LIGHT0, GL_POSITION, position);

  glEnable(GL_LIGHT0);
}


Color PointLight::illuminatedColor(const qglviewer::Vec& eye, const Hit& hit) const{
	Color res;

	qglviewer::Vec source = position_ - hit.intersection();
	source.normalize();

	float lidn = source*hit.normal();
	if(lidn<0)
		return Color(0,0,0);


	Color diffus;
	if(RTConfig::getInstance()->getTexEnable()){
		// "Partie XI: Textures"
		// With texture
		diffus = hit.material().diffuseColor(hit.u(), hit.v())*(lidn);
	}else{
		// Without texture
		diffus = hit.material().diffuseColor()*(lidn);
	}
	
	Color specular(0,0,0);

	//Phong
	if (RTConfig::getInstance()->getLightMode() == RTConfig::RTC_LM_PHONG) {
		qglviewer::Vec normale=hit.normal();
		qglviewer::Vec refl=-source+(normale*2.0f*(source*normale));

		float dp=eye*refl;
		if(dp<0)
			dp=0;

		specular = hit.material().specularColor() * pow( dp, hit.material().specularCoefficient());
	}

	//Blinn-Phong
	else {
		qglviewer::Vec h=(eye+source);
		h.normalize();

		float dp = h*hit.normal();
		if(dp<0)
			dp=0;

		specular = hit.material().specularColor() * powf(dp, hit.material().specularCoefficient());
	}


	res = color() * (diffus+specular);

	return res;
}


bool PointLight::visibleFrom(const qglviewer::Vec& pos, const Scene* const scene, Hit &rayhit) const{
	Ray ray;
//	Hit rayhit;

	ray.setStart(pos);
	ray.setDirection((position_-pos));

	rayhit.setTime(1.0E5);
	rayhit.setIntersection(qglviewer::Vec(0,0,0));

	scene->intersect(ray, rayhit);

	// return true only if there is no intersection 
	// or if the intersection is done behind the source (ie not between the source and the object)
	return (rayhit.time()==1.0E5 || (rayhit.time()>(position_-pos).norm()) );
}
